package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.mojang.datafixers.util.Either
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.data.DecimalCodec
import java.util.Optional

class ComputeAction(
	id: Either<ResourceLocation, TagKey<Item>>,
	values: Collection<Source>,
	priority: Optional<Int> = Optional.empty(),
) : AbstractRegistryAction(id, priority = priority) {
	companion object : Type<ComputeAction> {
		private val constantCodecConcise: Codec<Constant> by lazy {
			RecordCodecBuilder.create {
				it.group(
					DecimalCodec.fieldOf("matter").forGetter(Constant::matter),
					Codec.DOUBLE.fieldOf("complexity").forGetter(Constant::complexity),
					IMatterFunction.registry.codec.fieldOf("function").forGetter(Constant::matterFunction),
				).apply(it, ::Constant)
			}
		}

		private val constantCodecVeryConcise: Codec<Constant> by lazy {
			RecordCodecBuilder.create {
				it.group(
					DecimalCodec.fieldOf("value").forGetter(Constant::matter),
					IMatterFunction.registry.codec.fieldOf("function").forGetter(Constant::matterFunction),
				).apply(it, ::Constant)
			}
		}

		private val constantCodecVeryConcise2: Codec<Constant> by lazy {
			RecordCodecBuilder.create {
				it.group(
					Codec.DOUBLE.fieldOf("value").forGetter(Constant::complexity),
					IMatterFunction.registry.codec.fieldOf("function").forGetter(Constant::matterFunction),
				).apply(it, ::Constant)
			}
		}

		private val constantCodecVerbose: Codec<Constant> by lazy {
			RecordCodecBuilder.create {
				it.group(
					DecimalCodec.fieldOf("matter").forGetter(Constant::matter),
					Codec.DOUBLE.fieldOf("complexity").forGetter(Constant::complexity),
					IMatterFunction.registry.codec.fieldOf("matterFunction").forGetter(Constant::matterFunction),
					IMatterFunction.registry.codec.fieldOf("complexityFunction").forGetter(Constant::complexityFunction),
				).apply(it, ::Constant)
			}
		}

		private val constantCodec: Codec<Constant> by lazy {
			Codec.either(constantCodecConcise, constantCodecVerbose)
				.xmap({ it.map({ it }, { it }) }, { if (it.matterFunction == it.complexityFunction) Either.left(it) else Either.right(it) })
		}

		private object ActualConstantCodec : Codec<Constant> {
			override fun <T : Any> encode(input: Constant, ops: DynamicOps<T>, prefix: T): DataResult<T> {
				if (input.complexity == input.matter.toDouble()) {
					return constantCodecVeryConcise.encode(input, ops, prefix)
				} else {
					return constantCodec.encode(input, ops, prefix)
				}
			}

			override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<Constant, T>> {
				var result = constantCodecVeryConcise.decode(ops, input)

				if (result.result().isPresent) {
					return result
				}

				result = constantCodecVeryConcise2.decode(ops, input)

				if (result.result().isPresent) {
					return result
				}

				return constantCodec.decode(ops, input)
			}
		}

		private val valueCodecConcise: Codec<Value> by lazy {
			RecordCodecBuilder.create {
				it.group(
					TargetCodec.fieldOf("id").forGetter(Value::id),
					IMatterFunction.registry.codec.fieldOf("function").forGetter(Value::matterFunction),
				).apply(it, ::Value)
			}
		}

		private val valueCodecVerbose: Codec<Value> by lazy {
			RecordCodecBuilder.create {
				it.group(
					TargetCodec.fieldOf("id").forGetter(Value::id),
					IMatterFunction.registry.codec.fieldOf("matterFunction").forGetter(Value::matterFunction),
					IMatterFunction.registry.codec.fieldOf("complexityFunction").forGetter(Value::complexityFunction),
				).apply(it, ::Value)
			}
		}

		private val valueCodec: Codec<Value> by lazy {
			Codec.either(valueCodecConcise, valueCodecVerbose)
				.xmap({ it.map({ it }, { it }) }, { if (it.matterFunction == it.complexityFunction) Either.left(it) else Either.right(it) })
		}

		private object ActualValueCodec : Codec<Value> {
			override fun <T : Any> encode(input: Value, ops: DynamicOps<T>, prefix: T): DataResult<T> {
				if (input.matterFunction == input.complexityFunction && input.matterFunction == IMatterFunction.PLUS) {
					return TargetCodec.encode(input.id, ops, prefix)
				} else {
					return valueCodec.encode(input, ops, prefix)
				}
			}

			override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<Value, T>> {
				val result = TargetCodec.decode(ops, input)

				if (result.result().isPresent) {
					return result.map { it.mapFirst { Value(it, IMatterFunction.PLUS) } }
				}

				return valueCodec.decode(ops, input)
			}
		}

		override val codec: Codec<ComputeAction> by lazy {
			RecordCodecBuilder.create {
				it.group(
					TargetCodec.fieldOf("id").forGetter(ComputeAction::id),
					Codec.list(Codec
						.either(ActualConstantCodec, ActualValueCodec)
						.xmap({ it.map({ it }, { it }) }, { if (it is Constant) Either.left(it) else Either.right(it as Value) }))
						.fieldOf("values")
						.forGetter(ComputeAction::values),
					Codec.INT.optionalFieldOf("prority").forGetter(ComputeAction::priority)
				).apply(it, ::ComputeAction)
			}
		}
	}

	override val type: Type<*>
		get() = Companion

	sealed class Source(
		val matterFunction: IMatterFunction,
		val complexityFunction: IMatterFunction
	) {
		internal abstract fun provideValues(): MatterManager.Result

		internal fun update(self: IMatterValue): MatterManager.Result {
			val grab = provideValues()

			if (grab.value == null) {
				return grab
			}

			return MatterManager.Result(MatterValue(
				matterFunction.updateValue(self.matter, grab.value.matter),
				complexityFunction.updateValue(self.complexity, grab.value.complexity)
			))
		}
	}

	class Constant(
		val matter: Decimal,
		val complexity: Double,
		matterFunction: IMatterFunction,
		complexityFunction: IMatterFunction = matterFunction,
	) : Source(matterFunction, complexityFunction) {
		constructor(
			value: Decimal,
			matterFunction: IMatterFunction,
			complexityFunction: IMatterFunction = matterFunction
		) : this(value, value.toDouble(), matterFunction, complexityFunction)

		constructor(
			value: Double,
			matterFunction: IMatterFunction,
			complexityFunction: IMatterFunction = matterFunction
		) : this(Decimal(value.toString()), value, matterFunction, complexityFunction)

		override fun provideValues(): MatterManager.Result {
			return MatterManager.Result(MatterValue(matter, complexity))
		}
	}

	class Value(
		val id: Either<ResourceLocation, TagKey<Item>>,
		matterFunction: IMatterFunction,
		complexityFunction: IMatterFunction = matterFunction,
	) : Source(matterFunction, complexityFunction) {
		override fun provideValues(): MatterManager.Result {
			return id.map(
				{
					MatterManager.compute(ForgeRegistries.ITEMS.getValue(it) ?: throw NullPointerException("$it does not point to anything"))
				},
				{
					MatterManager.compute(it)
				}
			)
		}
	}

	val values: ImmutableList<Source> = ImmutableList.copyOf(values)

	internal fun tryCompute(): MatterManager.Result {
		var compute: IMatterValue = IMatterValue.Companion

		for (fn in values) {
			val result = fn.update(compute)

			if (result.value == null) {
				return result
			} else {
				compute = result.value
			}
		}

		return MatterManager.Result(compute)
	}

	override fun doUpdate(registry: MutableCollection<IMutableRegistryEntry>, source: ResourceLocation) {
		// do nothing
	}
}
